import numpy as np
from pymoo.algorithms.base.genetic import GeneticAlgorithm
from pymoo.algorithms.soo.nonconvex.ga import FitnessSurvival
from pymoo.core.replacement import ImprovementReplacement
from pymoo.operators.mutation.nom import NoMutation
from pymoo.operators.sampling.lhs import LHS
from pymoo.termination.default import DefaultSingleObjectiveTermination
from pymoo.util.display.single import SingleObjectiveOutput
from pymoo.core.infill import InfillCriterion
from pymoode.operators.des import DES
from pymoode.operators.dex import DEX, _validate_deprecated_repair
import warnings


# =========================================================================================================
# Implementation
# =========================================================================================================

class VariantDE(InfillCriterion):

    def __init__(self,
                 variant="DE/rand/1/bin",
                 CR=0.7,
                 F=(0.5, 1.0),
                 gamma=1e-4,
                 de_repair="bounce-back",
                 mutation=None,
                 **kwargs):
        """InfillCriterion class for Differential Evolution

        Parameters
        ----------
        variant : str, optional
            Differential evolution strategy. Must be a string in the format: "DE/selection/n/crossover", in which, n in an integer of number of difference vectors, and crossover is either 'bin' or 'exp'. Selection variants are:

                - 'ranked'
                - 'rand'
                - 'best'
                - 'current-to-best'
                - 'current-to-best'
                - 'current-to-rand'
                - 'rand-to-best'

            The selection strategy 'ranked' might be helpful to improve convergence speed without much harm to diversity. Defaults to 'DE/rand/1/bin'.

        CR : float, optional
            Crossover parameter. Defined in the range [0, 1]
            To reinforce mutation, use higher values. To control convergence speed, use lower values.

        F : iterable of float or float, optional
            Scale factor or mutation parameter. Defined in the range (0, 2]
            To reinforce exploration, use higher values; for exploitation, use lower values.

        gamma : float, optional
            Jitter deviation parameter. Should be in the range (0, 2). Defaults to 1e-4.

        de_repair : str, optional
            Repair of DE mutant vectors. Is either callable or one of:

                - 'bounce-back'
                - 'midway'
                - 'rand-init'
                - 'to-bounds'

            If callable, has the form fun(X, Xb, xl, xu) in which X contains mutated vectors including violations, Xb contains reference vectors for repair in feasible space, xl is a 1d vector of lower bounds, and xu a 1d vector of upper bounds.
            Defaults to 'bounce-back'.

        mutation : Mutation, optional
            Pymoo's mutation operator after crossover. Defaults to NoMutation().
        
        repair : Repair, optional
            Pymoo's repair operator after mutation. Defaults to NoRepair().
        """

        # Fix deprecated pm kwargs
        kwargs, mutation = _fix_deprecated_pm_kwargs(kwargs, mutation)

        # Default initialization of InfillCriterion
        super().__init__(eliminate_duplicates=None, **kwargs)

        # Parse the information from the string
        _, selection_variant, n_diff, crossover_variant, = variant.split("/")
        n_diffs = int(n_diff)

        # When "to" in variant there are more than 1 difference vectors
        if "-to-" in variant:
            n_diffs += 1

        # Define parent selection operator
        self.selection = DES(selection_variant)

        # Default value for F
        if F is None:
            F = (0.0, 1.0)

        # Define crossover strategy (DE mutation is included)
        self.crossover = DEX(variant=crossover_variant,
                             CR=CR,
                             F=F,
                             gamma=gamma,
                             n_diffs=n_diffs,
                             at_least_once=True,
                             de_repair=de_repair)

        # Define posterior mutation strategy and repair
        self.mutation = mutation if mutation is not None else NoMutation()

    def _do(self, problem, pop, n_offsprings, **kwargs):

        # Select parents including donor vector
        parents = self.selection(problem, pop, n_offsprings, self.crossover.n_parents, to_pop=True, **kwargs)

        # Perform mutation included in DEX and crossover
        off = self.crossover(problem, parents, **kwargs)

        # Perform posterior mutation and repair if passed
        off = self.mutation(problem, off)

        return off


class DE(GeneticAlgorithm):

    def __init__(self,
                 pop_size=100,
                 sampling=LHS(),
                 variant="DE/rand/1/bin",
                 CR=0.7,
                 F=(0.5, 1.0),
                 gamma=1e-4,
                 de_repair="bounce-back",
                 output=SingleObjectiveOutput(),
                 **kwargs):
        """
        Single-objective Differential Evolution proposed by Storn and Price (1997).

        Storn, R. & Price, K., 1997. Differential evolution–a simple and efficient heuristic for global optimization over continuous spaces. J. Glob. Optim., 11(4), pp. 341-359.

        Parameters
        ----------
        pop_size : int, optional
            Population size. Defaults to 100.

        sampling : Sampling, optional
            Sampling strategy of pymoo. Defaults to LHS().

        variant : str, optional
            Differential evolution strategy. Must be a string in the format: "DE/selection/n/crossover", in which, n in an integer of number of difference vectors, and crossover is either 'bin' or 'exp'. Selection variants are:

                - 'ranked'
                - 'rand'
                - 'best'
                - 'current-to-best'
                - 'current-to-best'
                - 'current-to-rand'
                - 'rand-to-best'

            The selection strategy 'ranked' might be helpful to improve convergence speed without much harm to diversity. Defaults to 'DE/rand/1/bin'.

        CR : float, optional
            Crossover parameter. Defined in the range [0, 1]
            To reinforce mutation, use higher values. To control convergence speed, use lower values.

        F : iterable of float or float, optional
            Scale factor or mutation parameter. Defined in the range (0, 2]
            To reinforce exploration, use higher values; for exploitation, use lower values.

        gamma : float, optional
            Jitter deviation parameter. Should be in the range (0, 2). Defaults to 1e-4.

        de_repair : str, optional
            Repair of DE mutant vectors. Is either callable or one of:

                - 'bounce-back'
                - 'midway'
                - 'rand-init'
                - 'to-bounds'

            If callable, has the form fun(X, Xb, xl, xu) in which X contains mutated vectors including violations, Xb contains reference vectors for repair in feasible space, xl is a 1d vector of lower bounds, and xu a 1d vector of upper bounds.
            Defaults to 'bounce-back'.

        mutation : Mutation, optional
            Pymoo's mutation operator after crossover. Defaults to NoMutation().

        repair : Repair, optional
            Pymoo's repair operator after mutation. Defaults to NoRepair().
        """

        # Reapir old argument
        de_repair, kwargs["repair"] = _validate_deprecated_repair(de_repair, **kwargs)

        # Mating
        mating = VariantDE(variant=variant,
                           CR=CR,
                           F=F,
                           gamma=gamma,
                           de_repair=de_repair,
                           **kwargs)

        # Number of offsprings at each generation
        n_offsprings = pop_size

        super().__init__(pop_size=pop_size,
                         sampling=sampling,
                         mating=mating,
                         n_offsprings=n_offsprings,
                         eliminate_duplicates=None,
                         output=output,
                         **kwargs)

        self.termination = DefaultSingleObjectiveTermination()

    def _initialize_advance(self, infills=None, **kwargs):
        self.pop = FitnessSurvival().do(self.problem, infills, n_survive=self.pop_size)

    def _infill(self):

        infills = self.mating(self.problem, self.pop, self.n_offsprings, algorithm=self)

        return infills

    def _advance(self, infills=None, **kwargs):

        assert infills is not None, "This algorithms uses the AskAndTell interface thus infills must be provided."

        # One-to-one replacement survival
        self.pop = ImprovementReplacement().do(self.problem, self.pop, infills)

        # Sort the population by fitness to make the selection simpler for mating (not an actual survival, just sorting)
        self.pop = FitnessSurvival().do(self.problem, self.pop)

        # Set ranks
        self.pop.set("rank", np.arange(self.pop_size))


def _fix_deprecated_pm_kwargs(kwargs, mutation):
    if "pm" in kwargs:
        warnings.warn(
                    "pm is deprecated; use 'mutation' for compatibility purposes with pymoo",
                    DeprecationWarning, 2
                )
        if mutation is None:
            mutation = kwargs["pm"]

    return kwargs, mutation



import numpy as np
from pymoo.algorithms.moo.nsga3 import ReferenceDirectionSurvival
from pymoo.operators.sampling.lhs import LHS
from pymoo.util.misc import has_feasible
from pymoode.algorithms._nsde import NSDE
from pymoode.operators.dex import _validate_deprecated_repair

# =========================================================================================================
# Implementation
# =========================================================================================================


class NSDER(NSDE):

    def __init__(self,
                 ref_dirs,
                 pop_size=100,
                 variant="DE/rand/1/bin",
                 CR=0.7,
                 F=None,
                 gamma=1e-4,
                 **kwargs):
        """
        NSDE-R is an extension of NSDE to many-objective problems (Reddy & Dulikravich, 2019) using NSGA-III survival.

        S. R. Reddy and G. S. Dulikravich, "Many-objective differential evolution optimization based on reference points: NSDE-R," Struct. Multidisc. Optim., vol. 60, pp. 1455-1473, 2019.

        Parameters
        ----------
        ref_dirs : array like
            The reference directions that should be used during the optimization.

        pop_size : int, optional
            Population size. Defaults to 100.

        sampling : Sampling, optional
            Sampling strategy of pymoo. Defaults to LHS().

        variant : str, optional
            Differential evolution strategy. Must be a string in the format: "DE/selection/n/crossover", in which, n in an integer of number of difference vectors, and crossover is either 'bin' or 'exp'. Selection variants are:

                - "ranked'
                - 'rand'
                - 'best'
                - 'current-to-best'
                - 'current-to-best'
                - 'current-to-rand'
                - 'rand-to-best'

            The selection strategy 'ranked' might be helpful to improve convergence speed without much harm to diversity. Defaults to 'DE/rand/1/bin'.

        CR : float, optional
            Crossover parameter. Defined in the range [0, 1]
            To reinforce mutation, use higher values. To control convergence speed, use lower values.

        F : iterable of float or float, optional
            Scale factor or mutation parameter. Defined in the range (0, 2]
            To reinforce exploration, use higher values; for exploitation, use lower values.

        gamma : float, optional
            Jitter deviation parameter. Should be in the range (0, 2). Defaults to 1e-4.

        de_repair : str, optional
            Repair of DE mutant vectors. Is either callable or one of:

                - 'bounce-back'
                - 'midway'
                - 'rand-init'
                - 'to-bounds'

            If callable, has the form fun(X, Xb, xl, xu) in which X contains mutated vectors including violations, Xb contains reference vectors for repair in feasible space, xl is a 1d vector of lower bounds, and xu a 1d vector of upper bounds.
            Defaults to 'bounce-back'.

        mutation : Mutation, optional
            Pymoo's mutation operator after crossover. Defaults to NoMutation().

        repair : Repair, optional
            Pymoo's repair operator after mutation. Defaults to NoRepair().

        survival : Survival, optional
            Pymoo's survival strategy.
            Defaults to ReferenceDirectionSurvival().
        """

        self.ref_dirs = ref_dirs

        if self.ref_dirs is not None:

            if pop_size is None:
                pop_size = len(self.ref_dirs)

            if pop_size < len(self.ref_dirs):
                print(
                    f"WARNING: pop_size={pop_size} is less than the number of reference directions ref_dirs={len(self.ref_dirs)}.\n"
                    "This might cause unwanted behavior of the algorithm. \n"
                    "Please make sure pop_size is equal or larger than the number of reference directions. ")

        if 'survival' in kwargs:
            survival = kwargs['survival']
            del kwargs['survival']
        else:
            survival = ReferenceDirectionSurvival(ref_dirs)

        super().__init__(pop_size=pop_size,
                         variant=variant,
                         CR=CR,
                         F=F,
                         gamma=gamma,
                         survival=survival,
                         **kwargs)

    def _setup(self, problem, **kwargs):

        if self.ref_dirs is not None:
            if self.ref_dirs.shape[1] != problem.n_obj:
                raise Exception(
                    "Dimensionality of reference points must be equal to the number of objectives: %s != %s" %
                    (self.ref_dirs.shape[1], problem.n_obj))

    def _set_optimum(self, **kwargs):
        if not has_feasible(self.pop):
            self.opt = self.pop[[np.argmin(self.pop.get("CV"))]]
        else:
            self.opt = self.survival.opt



from pymoode.algorithms._nsde import NSDE
from pymoo.core.population import Population
from pymoo.util.dominator import get_relation
from pymoode.survival import RankAndCrowding


# =========================================================================================================
# Implementation
# =========================================================================================================


class GDE3(NSDE):

    def __init__(self,
                 pop_size=100,
                 variant="DE/rand/1/bin",
                 CR=0.5,
                 F=None,
                 gamma=1e-4,
                 **kwargs):
        """
        GDE3 is an extension of DE to multi-objective problems using a mixed type survival strategy.
        It is implemented in this version with the same constraint handling strategy of NSGA-II by default.

        Derived algorithms GDE3-MNN and GDE3-2NN use by default survival RankAndCrowding with metrics 'mnn' and '2nn'.

        For many-objective problems, try using NSDE-R, GDE3-MNN, or GDE3-2NN.

        For Bi-objective problems, survival = RankAndCrowding(crowding_func='pcd') is very effective.

        Kukkonen, S. & Lampinen, J., 2005. GDE3: The third evolution step of generalized differential evolution. 2005 IEEE congress on evolutionary computation, Volume 1, pp. 443-450.

        Parameters
        ----------
        pop_size : int, optional
            Population size. Defaults to 100.

        sampling : Sampling, optional
            Sampling strategy of pymoo. Defaults to LHS().

        variant : str, optional
            Differential evolution strategy. Must be a string in the format: "DE/selection/n/crossover", in which, n in an integer of number of difference vectors, and crossover is either 'bin' or 'exp'. Selection variants are:

                - 'ranked'
                - 'rand'
                - 'best'
                - 'current-to-best'
                - 'current-to-best'
                - 'current-to-rand'
                - 'rand-to-best'

            The selection strategy 'ranked' might be helpful to improve convergence speed without much harm to diversity. Defaults to 'DE/rand/1/bin'.

        CR : float, optional
            Crossover parameter. Defined in the range [0, 1]
            To reinforce mutation, use higher values. To control convergence speed, use lower values.

        F : iterable of float or float, optional
            Scale factor or mutation parameter. Defined in the range (0, 2]
            To reinforce exploration, use higher values; for exploitation, use lower values.

        gamma : float, optional
            Jitter deviation parameter. Should be in the range (0, 2). Defaults to 1e-4.

        de_repair : str, optional
            Repair of DE mutant vectors. Is either callable or one of:

                - 'bounce-back'
                - 'midway'
                - 'rand-init'
                - 'to-bounds'

            If callable, has the form fun(X, Xb, xl, xu) in which X contains mutated vectors including violations, Xb contains reference vectors for repair in feasible space, xl is a 1d vector of lower bounds, and xu a 1d vector of upper bounds.
            Defaults to 'bounce-back'.

        mutation : Mutation, optional
            Pymoo's mutation operator after crossover. Defaults to NoMutation().

        repair : Repair, optional
            Pymoo's repair operator after mutation. Defaults to NoRepair().

        survival : Survival, optional
            Pymoo's survival strategy.
            Defaults to RankAndCrowding() with crowding distances ('cd').
            In GDE3, the survival strategy is applied after a one-to-one comparison between child vector and corresponding parent when both are non-dominated by the other.
        """

        super().__init__(pop_size=pop_size,
                         variant=variant,
                         CR=CR,
                         F=F,
                         gamma=gamma,
                         **kwargs)

    def _advance(self, infills=None, **kwargs):

        assert infills is not None, "This algorithms uses the AskAndTell interface thus 'infills' must to be provided."

        # The individuals that are considered for the survival later and final survive
        survivors = []

        # now for each of the infill solutions
        for k in range(len(self.pop)):

            # Get the offspring an the parent it is coming from
            off, parent = infills[k], self.pop[k]

            # Check whether the new solution dominates the parent or not
            rel = get_relation(parent, off)

            # If indifferent we add both
            if rel == 0:
                survivors.extend([parent, off])

            # If offspring dominates parent
            elif rel == -1:
                survivors.append(off)

            # If parent dominates offspring
            else:
                survivors.append(parent)

        # Create the population
        survivors = Population.create(*survivors)

        # Perform a survival to reduce to pop size
        self.pop = self.survival.do(self.problem, survivors, n_survive=self.n_offsprings)


class GDE3MNN(GDE3):

    def __init__(self, pop_size=100, variant="DE/rand/1/bin", CR=0.5, F=None, gamma=0.0001, **kwargs):
        survival = RankAndCrowding(crowding_func="mnn")
        super().__init__(pop_size, variant, CR, F, gamma, survival=survival, **kwargs)


class GDE32NN(GDE3):

    def __init__(self, pop_size=100, variant="DE/rand/1/bin", CR=0.5, F=None, gamma=0.0001, **kwargs):
        survival = RankAndCrowding(crowding_func="2nn")
        super().__init__(pop_size, variant, CR, F, gamma, survival=survival, **kwargs)


class GDE3PCD(GDE3):

    def __init__(self, pop_size=100, variant="DE/rand/1/bin", CR=0.5, F=None, gamma=0.0001, **kwargs):
        survival = RankAndCrowding(crowding_func="pcd")
        super().__init__(pop_size, variant, CR, F, gamma, survival=survival, **kwargs)



import numpy as np
import warnings
from pymoo.util.randomized_argsort import randomized_argsort
from pymoo.util.nds.non_dominated_sorting import NonDominatedSorting
from pymoo.core.survival import Survival, split_by_feasibility
from pymoo.core.population import Population
from pymoode.survival._metrics import get_crowding_function


class RankAndCrowding(Survival):

    def __init__(
        self,
        nds=None,
        crowding_func="cd",
        rule=None
    ):
        """
        A generalization of the NSGA-II survival operator that ranks individuals by dominance criteria
        and sorts the last front by some crowding metric.

        For many-objective problems, try using 'mnn' or '2nn'.

        For Bi-objective problems, 'pcd' is very effective.

        Parameters
        ----------
        nds : str or None, optional
            Pymoo type of non-dominated sorting. Defaults to None.

        crowding_func : str or callable, optional
            Crowding metric. Options are:

                - 'cd': crowding distances
                - 'pcd' or 'pruning-cd': improved pruning based on crowding distances
                - 'ce': crowding entropy
                - 'mnn': M-Neaest Neighbors
                - '2nn': 2-Neaest Neighbors

            If callable, it has the form ``fun(F, filter_out_duplicates=None, n_remove=None, **kwargs)``
            in which F (n, m) and must return metrics in a (n,) array.

            The options 'pcd', 'cd', and 'ce' are recommended for two-objective problems, whereas 'mnn' and '2nn' for many objective.
            When using 'pcd', 'mnn', or '2nn', individuals are already eliminated in a 'single' manner. 
            Due to Cython implementation, they are as fast as the corresponding 'cd', 'mnn-fast', or '2nn-fast', 
            although they can singnificantly improve diversity of solutions.
            Defaults to 'cd'.
        """
        if rule is not None:
            warnings.warn(
                "rule is deprecated; elimination methods are now included in crowding_func",
                DeprecationWarning, 2
            )

        crowding_func_ = get_crowding_function(crowding_func)

        super().__init__(filter_infeasible=True)
        self.nds = nds if nds is not None else NonDominatedSorting()
        self.crowding_func = crowding_func_

    def _do(self,
            problem,
            pop,
            *args,
            n_survive=None,
            **kwargs):

        # get the objective space values and objects
        F = pop.get("F").astype(float, copy=False)

        # the final indices of surviving individuals
        survivors = []

        # do the non-dominated sorting until splitting front
        fronts = self.nds.do(F, n_stop_if_ranked=n_survive)

        for k, front in enumerate(fronts):

            # current front sorted by crowding distance if splitting
            while len(survivors) + len(front) > n_survive:

                # Define how many will be removed
                n_remove = len(survivors) + len(front) - n_survive

                # re-calculate the crowding distance of the front
                crowding_of_front = \
                    self.crowding_func.do(
                        F[front, :],
                        n_remove=n_remove
                    )

                I = randomized_argsort(crowding_of_front, order='descending', method='numpy')

                I = I[:-n_remove]
                front = front[I]

            # otherwise take the whole front unsorted
            else:
                # calculate the crowding distance of the front
                crowding_of_front = \
                    self.crowding_func.do(
                        F[front, :],
                        n_remove=0
                    )

            # save rank and crowding in the individual class
            for j, i in enumerate(front):
                pop[i].set("rank", k)
                pop[i].set("crowding", crowding_of_front[j])

            # extend the survivors by all or selected individuals
            survivors.extend(front)

        return pop[survivors]


class ConstrRankAndCrowding(Survival):

    def __init__(self, nds=None, crowding_func="cd", ranking=None):
        """
        The Rank and Crowding survival approach for handling constraints proposed on
        GDE3 by Kukkonen, S. & Lampinen, J. (2005).

        Parameters
        ----------
        nds : str or None, optional
            Pymoo type of non-dominated sorting. Defaults to None.

        crowding_func : str or callable, optional
            Crowding metric. Options are:

                - 'cd': crowding distances
                - 'pcd' or 'pruning-cd': improved pruning based on crowding distances
                - 'ce': crowding entropy
                - 'mnn': M-Neaest Neighbors
                - '2nn': 2-Neaest Neighbors

            If callable, it has the form ``fun(F, filter_out_duplicates=None, n_remove=None, **kwargs)``
            in which F (n, m) and must return metrics in a (n,) array.

            The options 'pcd', 'cd', and 'ce' are recommended for two-objective problems, whereas 'mnn' and '2nn' for many objective.
            When using 'pcd', 'mnn', or '2nn', individuals are already eliminated in a 'single' manner. 
            Due to Cython implementation, they are as fast as the corresponding 'cd', 'mnn-fast', or '2nn-fast', 
            although they can singnificantly improve diversity of solutions.
            Defaults to 'cd'.
        """
        if ranking is not None:
            warnings.warn(
                "'ranking' is deprecated; internal survival operators are now constructed from nds and crowding_func",
                DeprecationWarning, 2
            )

        super().__init__(filter_infeasible=False)
        self.nds = nds if nds is not None else NonDominatedSorting()
        self.ranking = RankAndCrowding(nds=nds, crowding_func=crowding_func)

    def _do(self,
            problem,
            pop,
            *args,
            n_survive=None,
            **kwargs):

        if n_survive is None:
            n_survive = len(pop)

        n_survive = min(n_survive, len(pop))

        # If the split should be done beforehand
        if problem.n_constr > 0:

            # Split by feasibility
            feas, infeas = split_by_feasibility(pop, sort_infeas_by_cv=True, sort_feas_by_obj=False, return_pop=False)

            # Obtain len of feasible
            n_feas = len(feas)

            # Assure there is at least_one survivor
            if n_feas == 0:
                survivors = Population()
            else:
                survivors = self.ranking.do(problem, pop[feas], *args, n_survive=min(len(feas), n_survive), **kwargs)

            # Calculate how many individuals are still remaining to be filled up with infeasible ones
            n_remaining = n_survive - len(survivors)

            # If infeasible solutions need to be added
            if n_remaining > 0:

                # Constraints to new ranking
                G = pop[infeas].get("G")
                G = np.maximum(G, 0)

                # Fronts in infeasible population
                infeas_fronts = self.nds.do(G, n_stop_if_ranked=n_remaining)

                # Iterate over fronts
                for k, front in enumerate(infeas_fronts):

                    # Save ranks
                    pop[infeas][front].set("cv_rank", k)

                    # Current front sorted by CV
                    if len(survivors) + len(front) > n_survive:

                        # Obtain CV of front
                        CV = pop[infeas][front].get("CV").flatten()
                        I = randomized_argsort(CV, order='ascending', method='numpy')
                        I = I[:(n_survive - len(survivors))]

                    # Otherwise take the whole front unsorted
                    else:
                        I = np.arange(len(front))

                    # extend the survivors by all or selected individuals
                    survivors = Population.merge(survivors, pop[infeas][front[I]])

        else:
            survivors = self.ranking.do(problem, pop, *args, n_survive=n_survive, **kwargs)

        return survivors



from pymoo.algorithms.moo.nsga2 import NSGA2
from pymoo.operators.sampling.lhs import LHS
from pymoode.algorithms._de import VariantDE
from pymoode.survival._classes import RankAndCrowding
from pymoode.operators.dex import _validate_deprecated_repair


# =========================================================================================================
# Implementation
# =========================================================================================================


class NSDE(NSGA2):

    def __init__(self,
                 pop_size=100,
                 sampling=LHS(),
                 variant="DE/rand/1/bin",
                 CR=0.7,
                 F=None,
                 gamma=1e-4,
                 de_repair="bounce-back",
                 survival=RankAndCrowding(),
                 **kwargs):
        """
        NSDE is an algorithm that combines that combines NSGA-II sorting and survival strategies 
        to DE mutation and crossover.

        For many-objective problems, try using NSDE-R, GDE3-MNN, or GDE3-2NN.

        For Bi-objective problems, survival = RankAndCrowding(crowding_func='pcd') is very effective.

        Parameters
        ----------
        pop_size : int, optional
            Population size. Defaults to 100.

        sampling : Sampling, optional
            Sampling strategy of pymoo. Defaults to LHS().

        variant : str, optional
            Differential evolution strategy. Must be a string in the format: "DE/selection/n/crossover", in which, n in an integer of number of difference vectors, and crossover is either 'bin' or 'exp'. Selection variants are:

                - "ranked'
                - 'rand'
                - 'best'
                - 'current-to-best'
                - 'current-to-best'
                - 'current-to-rand'
                - 'rand-to-best'

            The selection strategy 'ranked' might be helpful to improve convergence speed without much harm to diversity. Defaults to 'DE/rand/1/bin'.

        CR : float, optional
            Crossover parameter. Defined in the range [0, 1]
            To reinforce mutation, use higher values. To control convergence speed, use lower values.

        F : iterable of float or float, optional
            Scale factor or mutation parameter. Defined in the range (0, 2]
            To reinforce exploration, use higher values; for exploitation, use lower values.

        gamma : float, optional
            Jitter deviation parameter. Should be in the range (0, 2). Defaults to 1e-4.

        de_repair : str, optional
            Repair of DE mutant vectors. Is either callable or one of:

                - 'bounce-back'
                - 'midway'
                - 'rand-init'
                - 'to-bounds'

            If callable, has the form fun(X, Xb, xl, xu) in which X contains mutated vectors including violations, Xb contains reference vectors for repair in feasible space, xl is a 1d vector of lower bounds, and xu a 1d vector of upper bounds.
            Defaults to 'bounce-back'.

        mutation : Mutation, optional
            Pymoo's mutation operator after crossover. Defaults to NoMutation().

        repair : Repair, optional
            Pymoo's repair operator after mutation. Defaults to NoRepair().

        survival : Survival, optional
            Pymoo's survival strategy.
            Defaults to RankAndCrowding() with crowding distances ('cd').
            In GDE3, the survival strategy is applied after a one-to-one comparison between child vector and corresponding parent when both are non-dominated by the other.
        """

        # Reapir old argument
        de_repair, kwargs["repair"] = _validate_deprecated_repair(de_repair, **kwargs)

        # Number of offsprings at each generation
        n_offsprings = pop_size

        # Mating
        mating = VariantDE(variant=variant,
                           CR=CR,
                           F=F,
                           gamma=gamma,
                           de_repair=de_repair,
                           **kwargs)

        # Init from pymoo's NSGA2
        super().__init__(pop_size=pop_size,
                         sampling=sampling,
                         mating=mating,
                         survival=survival,
                         eliminate_duplicates=None,
                         n_offsprings=n_offsprings,
                         **kwargs)



import numpy as np
from pymoo.core.crossover import Crossover
from pymoo.core.population import Population
from pymoo.core.repair import Repair
import warnings


# =========================================================================================================
# Implementation
# =========================================================================================================

class DEM(Crossover):

    def __init__(self,
                 F=None,
                 gamma=1e-4,
                 de_repair="bounce-back",
                 n_diffs=1,
                 **kwargs):
        """Differential Evolution mutation
        (Implemented as pymoo crossover)

        Parameters
        ----------
        F : iterable of float or float, optional
            Scale factor or mutation parameter. Defined in the range (0, 2]
            To reinforce exploration, use higher values; for exploitation, use lower values.

        gamma : float, optional
            Jitter deviation parameter. Should be in the range (0, 2). Defaults to 1e-4.

        de_repair : str, optional
            Repair of DE mutant vectors. Is either callable or one of:

                - 'bounce-back'
                - 'midway'
                - 'rand-init'
                - 'to-bounds'

            If callable, has the form fun(X, Xb, xl, xu) in which X contains mutated vectors including violations, Xb contains reference vectors for repair in feasible space, xl is a 1d vector of lower bounds, and xu a 1d vector of upper bounds.
            Defaults to 'bounce-back'.
        
        n_diffs : int, optional
            Number of difference vectors in operation, by default 1

        Raises
        ------
        KeyError
            Bad de_repair input
        """

        # Crossover basic structure
        super().__init__(1 + 2 * n_diffs, 1,  prob=1.0, **kwargs)

        # Default value for F
        if F is None:
            F = (0.0, 1.0)

        # Define which method will be used to generate F values
        if hasattr(F, "__iter__"):
            self.scale_factor = self._randomize_scale_factor
        else:
            self.scale_factor = self._scalar_scale_factor

        # Define which method will be used to generate F values
        if not hasattr(de_repair, "__call__"):
            try:
                de_repair = REPAIRS[de_repair]
            except:
                raise KeyError("Repair must be either callable or in " + str(list(REPAIRS.keys())))

        # Define which strategy of rotation will be used
        if gamma is None:
            self.get_diff = self._diff_simple
        else:
            self.get_diff = self._diff_jitter

        self.F = F
        self.gamma = gamma
        self.de_repair = de_repair

    def do(self, problem, pop, parents=None, **kwargs):
        """Perform Differential Evolution mutation

        Parameters
        ----------
        problem : Problem
            Optimization problem
        
        pop : Population
            Original parent population at a given generation
        
        parents : numpy.array (n_offsprings, n_select) of dtype int, optional
            Parent (indexes) to crossover, by default None

        Returns
        -------
        Population
            Offspring mutants V
        """

        # Convert pop if parents is not None
        if parents is not None:
            pop = pop[parents]

        # Get all X values for mutation parents
        Xr = np.swapaxes(pop, 0, 1).get("X")

        # Create mutation vectors
        V, diffs = self.de_mutation(Xr, return_differentials=True)

        # If the problem has boundaries to be considered
        if problem.has_bounds():

            # Do de_repair
            V = self.de_repair(V, Xr[0], *problem.bounds())

        return Population.new("X", V)

    def de_mutation(self, Xr, return_differentials=True):

        n_parents, n_matings, n_var = Xr.shape
        assert n_parents % 2 == 1, "For the differential an odd number of values need to be provided"

        # Build the pairs for the differentials
        pairs = (np.arange(n_parents - 1) + 1).reshape(-1, 2)

        # The differentials from each pair subtraction
        diffs = self.get_diffs(Xr, pairs, n_matings, n_var)

        # Add the difference vectors to the base vector
        V = Xr[0] + diffs

        if return_differentials:
            return V, diffs
        else:
            return V

    def _randomize_scale_factor(self, n_matings):
        return (self.F[0] + np.random.random(n_matings) * (self.F[1] - self.F[0]))

    def _scalar_scale_factor(self, n_matings):
        return np.full(n_matings, self.F)

    def _diff_jitter(self, F, Xi, Xj, n_matings, n_var):
        F = F[:, None] * (1 + self.gamma * (np.random.random((n_matings, n_var)) - 0.5))
        return F * (Xi - Xj)

    def _diff_simple(self, F, Xi, Xj, n_matings, n_var):
        return F[:, None] * (Xi - Xj)

    def get_diffs(self, Xr, pairs, n_matings, n_var):

        # The differentials from each pair subtraction
        diffs = np.zeros((n_matings, n_var))

        # For each difference
        for i, j in pairs:

            # Obtain F randomized in range
            F = self.scale_factor(n_matings)

            # New difference vector
            diff = self.get_diff(F, Xr[i], Xr[j], n_matings, n_var)

            # Add the difference to the first vector
            diffs = diffs + diff

        return diffs


class DEX(Crossover):

    def __init__(self,
                 variant="bin",
                 CR=0.7,
                 F=None,
                 gamma=1e-4,
                 n_diffs=1,
                 at_least_once=True,
                 de_repair="bounce-back",
                 **kwargs):
        """Differential evolution crossover
        (DE mutation is considered a part of this operator)

        Parameters
        ----------
        variant : str, optional
            Crossover variant. Must be either "bin" or "exp". By default "bin"
        
        CR : float, optional
            Crossover parameter. Defined in the range [0, 1]
            To reinforce mutation, use higher values. To control convergence speed, use lower values.

        F : iterable of float or float, optional
            Scale factor or mutation parameter. Defined in the range (0, 2]
            To reinforce exploration, use higher values; for exploitation, use lower values.

        gamma : float, optional
            Jitter deviation parameter. Should be in the range (0, 2). Defaults to 1e-4.

        de_repair : str, optional
            Repair of DE mutant vectors. Is either callable or one of:

                - 'bounce-back'
                - 'midway'
                - 'rand-init'
                - 'to-bounds'

            If callable, has the form fun(X, Xb, xl, xu) in which X contains mutated vectors including violations, Xb contains reference vectors for repair in feasible space, xl is a 1d vector of lower bounds, and xu a 1d vector of upper bounds.
            Defaults to 'bounce-back'.
        """

        # Crossover basic structure
        super().__init__(2 + 2 * n_diffs, 1,  prob=1.0, **kwargs)

        # Default value for F
        if F is None:
            F = (0.0, 1.0)

        # Create instace for mutation
        self.dem = DEM(F=F,
                       gamma=gamma,
                       de_repair=de_repair,
                       n_diffs=n_diffs)

        self.CR = CR
        self.variant = variant
        self.at_least_once = at_least_once

    def do(self, problem, pop, parents=None, **kwargs):
        """Do Differential Evolution crossover

        Parameters
        ----------
        problem : Problem
            Optimization problem
        
        pop : Population
            Original parent population at a given generation
        
        parents : numpy.array (n_offsprings, n_select) of dtype int, optional
            Parent (indexes) to crossover, by default None

        Returns
        -------
        Population
            Offspring population

        Raises
        ------
        Exception
            Unknown variant
        """

        # Convert pop if parents is not None
        if parents is not None:
            pop = pop[parents]

        # Get all X values for mutation parents
        X = pop[:, 0].get("X")

        # About Xi
        n_matings, n_var = X.shape

        # Obtain mutants
        mutants = self.dem.do(problem, pop[:, 1:], **kwargs)

        # Obtain V
        V = mutants.get("X")

        # Binomial crossover
        if self.variant == "bin":
            M = mut_binomial(n_matings, n_var, self.CR, at_least_once=self.at_least_once)
        # Exponential crossover
        elif self.variant == "exp":
            M = mut_exp(n_matings, n_var, self.CR, at_least_once=self.at_least_once)
        else:
            raise Exception(f"Unknown variant: {self.variant}")

        # Add mutated elements in corresponding main parent
        X[M] = V[M]

        off = Population.new("X", X)

        return off


# =========================================================================================================
# Crossovers
# =========================================================================================================

# From version 0.5.0 of pymoo
def row_at_least_once_true(M):

    _, d = M.shape

    for k in np.where(~np.any(M, axis=1))[0]:
        M[k, np.random.randint(d)] = True

    return M


def mut_binomial(n_matings, n_var, prob, at_least_once=True):

    M = np.random.random((n_matings, n_var)) < prob

    if at_least_once:
        M = row_at_least_once_true(M)

    return M


def mut_exp(n_matings, n_var, prob, at_least_once=True):

    # the mask do to the crossover
    M = np.full((n_matings, n_var), False)

    # start point of crossover
    s = np.random.randint(0, n_var, size=n_matings)

    # create for each individual the crossover range
    for i in range(n_matings):

        # the actual index where we start
        start = s[i]
        for j in range(n_var):

            # the current position where we are pointing to
            current = (start + j) % n_var

            # replace only if random value keeps being smaller than CR
            if np.random.random() <= prob:
                M[i, current] = True
            else:
                break

    if at_least_once:
        M = row_at_least_once_true(M)

    return M


def _validate_deprecated_repair(de_repair, **kwargs):

    if "repair" in kwargs:

        repair = kwargs["repair"]
        if (repair in REPAIRS) or (hasattr(repair, "__call__") and (not isinstance(repair, Repair))):
            warnings.warn(
                    "repair is deprecated; DE repair methods are now included in de_repair argument",
                    DeprecationWarning, 2
                )
            de_repair = repair
            repair = None

        else:
            pass

    else:
        repair = None

    return de_repair, repair


# =========================================================================================================
# Repairs
# =========================================================================================================


def bounce_back(X, Xb, xl, xu):
    """Repair strategy that radomly re-initializes violated elements between reference points and bounds

    Parameters
    ----------
    X : numpy.array (n_samples, n_var)
        Original population in decision space (including violations)
    
    Xb : numpy.array (n_samples, n_var)
        Feasible reference points in decision space
    
    xl : numpy.array (n_var,)
        Lower bounds for decision variables
    
    xu : numpy.array (n_var,)
        Upper bounds for decision variables

    Returns
    -------
    numpy.array (n_samples, n_var)
        Repaired population
    """

    XL = xl[None, :].repeat(len(X), axis=0)
    XU = xu[None, :].repeat(len(X), axis=0)

    i, j = np.where(X < XL)
    if len(i) > 0:
        X[i, j] = XL[i, j] + np.random.random(len(i)) * (Xb[i, j] - XL[i, j])

    i, j = np.where(X > XU)
    if len(i) > 0:
        X[i, j] = XU[i, j] - np.random.random(len(i)) * (XU[i, j] - Xb[i, j])

    return X


def midway(X, Xb, xl, xu):
    """Repair strategy that sets violated elements to midpoint between reference points and bounds

    Parameters
    ----------
    X : numpy.array (n_samples, n_var)
        Original population in decision space (including violations)
    
    Xb : numpy.array (n_samples, n_var)
        Feasible reference points in decision space
    
    xl : numpy.array (n_var,)
        Lower bounds for decision variables
    
    xu : numpy.array (n_var,)
        Upper bounds for decision variables

    Returns
    -------
    numpy.array (n_samples, n_var)
        Repaired population
    """

    XL = xl[None, :].repeat(len(X), axis=0)
    XU = xu[None, :].repeat(len(X), axis=0)

    i, j = np.where(X < XL)
    if len(i) > 0:
        X[i, j] = XL[i, j] + (Xb[i, j] - XL[i, j]) / 2

    i, j = np.where(X > XU)
    if len(i) > 0:
        X[i, j] = XU[i, j] - (XU[i, j] - Xb[i, j]) / 2

    return X


def to_bounds(X, Xb, xl, xu):
    """Repair strategy that forces violated elements to problem bounds

    Parameters
    ----------
    X : numpy.array (n_samples, n_var)
        Original population in decision space (including violations)
    
    Xb : numpy.array (n_samples, n_var)
        Feasible reference points in decision space
    
    xl : numpy.array (n_var,)
        Lower bounds for decision variables
    
    xu : numpy.array (n_var,)
        Upper bounds for decision variables

    Returns
    -------
    numpy.array (n_samples, n_var)
        Repaired population
    """

    xl = np.array(xl)
    xu = np.array(xu)

    XL = xl[None, :].repeat(len(X), axis=0)
    XU = xu[None, :].repeat(len(X), axis=0)

    i, j = np.where(X < XL)
    if len(i) > 0:
        X[i, j] = XL[i, j]

    i, j = np.where(X > XU)
    if len(i) > 0:
        X[i, j] = XU[i, j]

    return X


def rand_init(X, Xb, xl, xu):
    """Repair strategy that radomly re-initializes violated elements to problem domain

    Parameters
    ----------
    X : numpy.array (n_samples, n_var)
        Original population in decision space (including violations)
    
    Xb : numpy.array (n_samples, n_var)
        Feasible reference points in decision space
    
    xl : numpy.array (n_var,)
        Lower bounds for decision variables
    
    xu : numpy.array (n_var,)
        Upper bounds for decision variables

    Returns
    -------
    numpy.array (n_samples, n_var)
        Repaired population
    """

    XL = xl[None, :].repeat(len(X), axis=0)
    XU = xu[None, :].repeat(len(X), axis=0)

    i, j = np.where(X < XL)
    if len(i) > 0:
        X[i, j] = XL[i, j] + np.random.random(len(i)) * (XU[i, j] - XL[i, j])

    i, j = np.where(X > XU)
    if len(i) > 0:
        X[i, j] = XU[i, j] - np.random.random(len(i)) * (XU[i, j] - XL[i, j])

    return X


def squared_bounce_back(X, Xb, xl, xu):
    """Repair strategy that radomly re-initializes violated elements between reference points and bounds
    closer to bounds

    Parameters
    ----------
    X : numpy.array (n_samples, n_var)
        Original population in decision space (including violations)
    
    Xb : numpy.array (n_samples, n_var)
        Feasible reference points in decision space
    
    xl : numpy.array (n_var,)
        Lower bounds for decision variables
    
    xu : numpy.array (n_var,)
        Upper bounds for decision variables

    Returns
    -------
    numpy.array (n_samples, n_var)
        Repaired population
    """

    XL = xl[None, :].repeat(len(X), axis=0)
    XU = xu[None, :].repeat(len(X), axis=0)

    i, j = np.where(X < XL)
    if len(i) > 0:
        X[i, j] = XL[i, j] + np.square(np.random.random(len(i))) * (Xb[i, j] - XL[i, j])

    i, j = np.where(X > XU)
    if len(i) > 0:
        X[i, j] = XU[i, j] - np.square(np.random.random(len(i))) * (XU[i, j] - Xb[i, j])

    return X


def normalize_fun(fun):

    fmin = fun.min(axis=0)
    fmax = fun.max(axis=0)
    den = fmax - fmin

    den[den <= 1e-16] = 1.0

    return (fun - fmin)/den


REPAIRS = {"bounce-back": bounce_back,
           "midway": midway,
           "rand-init": rand_init,
           "to-bounds": to_bounds}
